/*****************************************************************************
 * dbus.c : D-Bus control interface
 *****************************************************************************
 * Copyright © 2006-2008 Rafaël Carré
 * Copyright © 2007-2008 Mirsal Ennaime
 * Copyright © 2009 The VideoLAN team
 * Copyright © 2009 M2X BV
 *
 * $Id: 410ff4357e61d61fc8ca543b260c10bcfd490c18 $
 *
 * Authors:    Rafaël Carré <funman at videolanorg>
 *             Mirsal Ennaime <mirsal dot ennaime at gmail dot com>
 *             Jean-Paul Saman <jean-paul dot saman at m2x dot nl>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston MA 02110-1301, USA.
 *****************************************************************************/

/*
 * D-Bus Specification:
 *      http://dbus.freedesktop.org/doc/dbus-specification.html
 * D-Bus low-level C API (libdbus)
 *      http://dbus.freedesktop.org/doc/dbus/api/html/index.html
 *  extract:
 *   "If you use this low-level API directly, you're signing up for some pain."
 *
 * MPRIS Specification (still drafting on Jan, 23 of 2008):
 *      http://wiki.xmms2.xmms.se/index.php/MPRIS
 */

/*****************************************************************************
 * Preamble
 *****************************************************************************/

#ifdef HAVE_CONFIG_H
# include "config.h"
#endif

#include <dbus/dbus.h>
#include "dbus.h"

#include <vlc_common.h>
#include <vlc_plugin.h>
#include <vlc_aout.h>
#include <vlc_vout.h>
#include <vlc_interface.h>
#include <vlc_playlist.h>

#include <vlc_iso_lang.h>

#include <math.h>

#include <assert.h>

/*****************************************************************************
 * Local prototypes.
 *****************************************************************************/

static int  Open    ( vlc_object_t * );
static void Close   ( vlc_object_t * );
static void Run     ( intf_thread_t * );

static int StateChange( intf_thread_t *, int );
static int TrackChange( intf_thread_t * );
static int StatusChangeEmit( intf_thread_t *);
static int TrackListChangeEmit( intf_thread_t *, int, int );

static int AllCallback( vlc_object_t*, const char*, vlc_value_t, vlc_value_t, void* );

static int GetInputMeta ( input_item_t *, DBusMessageIter * );
static int GetInputEsMeta( input_item_t *, int32_t, DBusMessageIter * );
static int MarshalStatus ( intf_thread_t *, DBusMessageIter *, bool );
static int UpdateCaps( intf_thread_t* );

/* GetCaps() capabilities */
enum
{
     CAPS_NONE                  = 0,      /* MPRIS v1.0 */
     CAPS_CAN_GO_NEXT           = 1 << 0,
     CAPS_CAN_GO_PREV           = 1 << 1,
     CAPS_CAN_PAUSE             = 1 << 2,
     CAPS_CAN_PLAY              = 1 << 3,
     CAPS_CAN_SEEK              = 1 << 4,
     CAPS_CAN_PROVIDE_METADATA  = 1 << 5,
     CAPS_CAN_HAS_TRACKLIST     = 1 << 6,
     CAPS_CAN_CONTROL_RATE      = 1 << 7,  /* MPRIS v2.0 */
     CAPS_CAN_CONTROL_REWIND    = 1 << 8,  /* MPRIS v2.0 */
     CAPS_CAN_CONTROL_SUBTITLES = 1 << 9,  /* MPRIS v2.0 */
     CAPS_CAN_CONTROL_TELETEXT  = 1 << 10  /* MPRIS v2.0 */
};

// The signal that can be get from the callbacks
enum
{
    SIGNAL_ITEM_CURRENT,
    SIGNAL_INTF_CHANGE,
    SIGNAL_PLAYLIST_ITEM_APPEND,
    SIGNAL_PLAYLIST_ITEM_DELETED,
    SIGNAL_RANDOM,
    SIGNAL_REPEAT,
    SIGNAL_LOOP,
    SIGNAL_STATE
};

struct intf_sys_t
{
    DBusConnection *p_conn;
    bool            b_meta_read;
    dbus_int32_t    i_caps;
    bool            b_dead;
    vlc_array_t    *p_events;
    vlc_mutex_t     lock;
};

typedef struct
{
    int signal;
    int i_node;
    int i_input_state;
} callback_info_t;

/*****************************************************************************
 * Module descriptor
 *****************************************************************************/

vlc_module_begin ()
    set_shortname( N_("dbus"))
    set_category( CAT_INTERFACE )
    set_subcategory( SUBCAT_INTERFACE_CONTROL )
    set_description( N_("D-Bus control interface") )
    set_capability( "interface", 0 )
    set_callbacks( Open, Close )
vlc_module_end ()

/*****************************************************************************
 * Methods
 *****************************************************************************/

/* Player */

DBUS_METHOD( Quit )
{ /* exits vlc */
    REPLY_INIT;
    libvlc_Quit(((vlc_object_t*)p_this)->p_libvlc);
    REPLY_SEND;
}

DBUS_METHOD( MprisVersion )
{ /*implemented version of the mpris spec */
    REPLY_INIT;
    OUT_ARGUMENTS;
    VLC_UNUSED( p_this );
    dbus_uint16_t i_major = VLC_MPRIS_VERSION_MAJOR;
    dbus_uint16_t i_minor = VLC_MPRIS_VERSION_MINOR;
    DBusMessageIter version;

    if( !dbus_message_iter_open_container( &args, DBUS_TYPE_STRUCT, NULL,
            &version ) )
        return DBUS_HANDLER_RESULT_NEED_MEMORY;

    if( !dbus_message_iter_append_basic( &version, DBUS_TYPE_UINT16,
            &i_major ) )
        return DBUS_HANDLER_RESULT_NEED_MEMORY;

    if( !dbus_message_iter_append_basic( &version, DBUS_TYPE_UINT16,
            &i_minor ) )
        return DBUS_HANDLER_RESULT_NEED_MEMORY;

    if( !dbus_message_iter_close_container( &args, &version ) )
        return DBUS_HANDLER_RESULT_NEED_MEMORY;
    REPLY_SEND;
}

DBUS_METHOD( PositionGet )
{ /* returns position in milliseconds */
    REPLY_INIT;
    OUT_ARGUMENTS;
    vlc_value_t position;
    dbus_int32_t i_pos;

    playlist_t *p_playlist = pl_Hold( ((vlc_object_t*) p_this) );
    input_thread_t *p_input = playlist_CurrentInput( p_playlist );

    if( !p_input )
        i_pos = 0;
    else
    {
        var_Get( p_input, "time", &position );
        i_pos = position.i_time / 1000;
        vlc_object_release( p_input );
    }
    pl_Release( ((vlc_object_t*) p_this) );
    ADD_INT32( &i_pos );
    REPLY_SEND;
}

DBUS_METHOD( PositionSet )
{ /* set position in milliseconds */

    REPLY_INIT;
    vlc_value_t position;
    playlist_t* p_playlist = NULL;
    dbus_int32_t i_pos;

    DBusError error;
    dbus_error_init( &error );

    dbus_message_get_args( p_from, &error,
            DBUS_TYPE_INT32, &i_pos,
            DBUS_TYPE_INVALID );

    if( dbus_error_is_set( &error ) )
    {
        msg_Err( (vlc_object_t*) p_this, "D-Bus message reading : %s",
                error.message );
        dbus_error_free( &error );
        return DBUS_HANDLER_RESULT_NOT_YET_HANDLED;
    }
    p_playlist = pl_Hold( ((vlc_object_t*) p_this) );
    input_thread_t *p_input = playlist_CurrentInput( p_playlist );

    if( p_input )
    {
        position.i_time = ((mtime_t)i_pos) * 1000;
        var_Set( p_input, "time", position );
        vlc_object_release( p_input );
    }
    pl_Release( ((vlc_object_t*) p_this) );
    REPLY_SEND;
}

DBUS_METHOD( VolumeGet )
{ /* returns volume in percentage */
    REPLY_INIT;
    OUT_ARGUMENTS;
    dbus_int32_t i_dbus_vol;
    audio_volume_t i_vol;
    /* 2nd argument of aout_VolumeGet is int32 */
    aout_VolumeGet( (vlc_object_t*) p_this, &i_vol );
    double f_vol = 100. * i_vol / AOUT_VOLUME_MAX;
    i_dbus_vol = round( f_vol );
    ADD_INT32( &i_dbus_vol );
    REPLY_SEND;
}

DBUS_METHOD( VolumeSet )
{ /* set volume in percentage */
    REPLY_INIT;

    DBusError error;
    dbus_error_init( &error );

    dbus_int32_t i_dbus_vol;
    audio_volume_t i_vol;

    dbus_message_get_args( p_from, &error,
            DBUS_TYPE_INT32, &i_dbus_vol,
            DBUS_TYPE_INVALID );

    if( dbus_error_is_set( &error ) )
    {
        msg_Err( (vlc_object_t*) p_this, "D-Bus message reading : %s",
                error.message );
        dbus_error_free( &error );
        return DBUS_HANDLER_RESULT_NOT_YET_HANDLED;
    }

    double f_vol = AOUT_VOLUME_MAX * i_dbus_vol / 100.;
    i_vol = round( f_vol );
    aout_VolumeSet( (vlc_object_t*) p_this, i_vol );

    REPLY_SEND;
}

DBUS_METHOD( Next )
{ /* next playlist item */
    REPLY_INIT;
    playlist_t *p_playlist = pl_Hold( ((vlc_object_t*) p_this) );
    playlist_Next( p_playlist );
    pl_Release( ((vlc_object_t*) p_this) );
    REPLY_SEND;
}

DBUS_METHOD( Prev )
{ /* previous playlist item */
    REPLY_INIT;
    playlist_t *p_playlist = pl_Hold( ((vlc_object_t*) p_this) );
    playlist_Prev( p_playlist );
    pl_Release( ((vlc_object_t*) p_this) );
    REPLY_SEND;
}

DBUS_METHOD( Stop )
{ /* stop playing */
    REPLY_INIT;
    playlist_t *p_playlist = pl_Hold( ((vlc_object_t*) p_this) );
    playlist_Stop( p_playlist );
    pl_Release( ((vlc_object_t*) p_this) );
    REPLY_SEND;
}

DBUS_METHOD( GetStatus )
{ /* returns the current status as a struct of 4 ints */
/*
    First   0 = Playing, 1 = Paused, 2 = Stopped.
    Second  0 = Playing linearly , 1 = Playing randomly.
    Third   0 = Go to the next element once the current has finished playing , 1 = Repeat the current element
    Fourth  0 = Stop playing once the last element has been played, 1 = Never give up playing *
 */
    REPLY_INIT;
    OUT_ARGUMENTS;

    MarshalStatus( p_this, &args, true );

    REPLY_SEND;
}

DBUS_METHOD( Pause )
{
    REPLY_INIT;
    playlist_t *p_playlist = pl_Hold( (vlc_object_t*) p_this );
    playlist_Pause( p_playlist );
    pl_Release( (vlc_object_t*) p_this );
    REPLY_SEND;
}

DBUS_METHOD( Play )
{
    REPLY_INIT;
    playlist_t *p_playlist = pl_Hold( (vlc_object_t*) p_this );

    input_thread_t *p_input =  playlist_CurrentInput( p_playlist );

    if( p_input )
    {
        double i_pos = 0;
        input_Control( p_input, INPUT_SET_POSITION, i_pos );
        vlc_object_release( p_input );
    }
    else
        playlist_Play( p_playlist );

    pl_Release( (vlc_object_t*) p_this );
    REPLY_SEND;
}

DBUS_METHOD( GetCurrentMetadata )
{
    REPLY_INIT;
    OUT_ARGUMENTS;
    playlist_t* p_playlist = pl_Hold( (vlc_object_t*) p_this );
    PL_LOCK;
    playlist_item_t* p_item = playlist_CurrentPlayingItem( p_playlist );
    if( p_item )
        GetInputMeta( p_item->p_input, &args );
    PL_UNLOCK;
    pl_Release( (vlc_object_t*) p_this );
    REPLY_SEND;
}

DBUS_METHOD( GetCaps )
{
    REPLY_INIT;
    OUT_ARGUMENTS;

    ADD_INT32( &((intf_thread_t*)p_this)->p_sys->i_caps );

    REPLY_SEND;
}

/* Media Player information */

DBUS_METHOD( Identity )
{
    VLC_UNUSED(p_this);
    REPLY_INIT;
    OUT_ARGUMENTS;
    char *psz_identity;
    if( asprintf( &psz_identity, "%s %s", PACKAGE, VERSION ) != -1 )
    {
        ADD_STRING( &psz_identity );
        free( psz_identity );
    }
    else
        return DBUS_HANDLER_RESULT_NEED_MEMORY;

    REPLY_SEND;
}

/* TrackList */

DBUS_METHOD( AddTrack )
{ /* add the string to the playlist, and play it if the boolean is true */
    REPLY_INIT;
    OUT_ARGUMENTS;

    DBusError error;
    dbus_error_init( &error );
    playlist_t* p_playlist = NULL;

    char *psz_mrl;
    dbus_bool_t b_play;

    dbus_message_get_args( p_from, &error,
            DBUS_TYPE_STRING, &psz_mrl,
            DBUS_TYPE_BOOLEAN, &b_play,
            DBUS_TYPE_INVALID );

    if( dbus_error_is_set( &error ) )
    {
        msg_Err( (vlc_object_t*) p_this, "D-Bus message reading : %s",
                error.message );
        dbus_error_free( &error );
        return DBUS_HANDLER_RESULT_NOT_YET_HANDLED;
    }

    p_playlist = pl_Hold( (vlc_object_t*) p_this );
    playlist_Add( p_playlist, psz_mrl, NULL, PLAYLIST_APPEND |
            ( ( b_play == TRUE ) ? PLAYLIST_GO : 0 ) ,
            PLAYLIST_END, true, false );
    pl_Release( (vlc_object_t*) p_this );

    dbus_int32_t i_success = 0;
    ADD_INT32( &i_success );

    REPLY_SEND;
}

DBUS_METHOD( GetCurrentTrack )
{
    REPLY_INIT;
    OUT_ARGUMENTS;

    playlist_t *p_playlist = pl_Hold( (vlc_object_t*) p_this );
    dbus_int32_t i_position = p_playlist->i_current_index;
    pl_Release( (vlc_object_t*) p_this );

    ADD_INT32( &i_position );
    REPLY_SEND;
}

DBUS_METHOD( GetMetadata )
{
    REPLY_INIT;
    OUT_ARGUMENTS;
    DBusError error;
    dbus_error_init( &error );

    dbus_int32_t i_position;
    playlist_t *p_playlist;

    dbus_message_get_args( p_from, &error,
           DBUS_TYPE_INT32, &i_position,
           DBUS_TYPE_INVALID );

    if( dbus_error_is_set( &error ) )
    {
        msg_Err( (vlc_object_t*) p_this, "D-Bus message reading : %s",
                error.message );
        dbus_error_free( &error );
        return DBUS_HANDLER_RESULT_NOT_YET_HANDLED;
    }

    p_playlist = pl_Hold( (vlc_object_t*) p_this );
    PL_LOCK;
    if( i_position < p_playlist->current.i_size )
    {
        GetInputMeta( p_playlist->current.p_elems[i_position]->p_input, &args );
    }

    PL_UNLOCK;
    pl_Release( (vlc_object_t*) p_this );
    REPLY_SEND;
}

DBUS_METHOD( GetLength )
{
    REPLY_INIT;
    OUT_ARGUMENTS;

    playlist_t *p_playlist = pl_Hold( (vlc_object_t*) p_this );
    PL_LOCK;
    dbus_int32_t i_elements = playlist_CurrentSize( p_playlist );
    PL_UNLOCK;
    pl_Release( (vlc_object_t*) p_this );

    ADD_INT32( &i_elements );
    REPLY_SEND;
}

DBUS_METHOD( DelTrack )
{
    REPLY_INIT;

    DBusError error;
    dbus_error_init( &error );

    dbus_int32_t i_position;
    playlist_t *p_playlist;

    dbus_message_get_args( p_from, &error,
            DBUS_TYPE_INT32, &i_position,
            DBUS_TYPE_INVALID );

    if( dbus_error_is_set( &error ) )
    {
        msg_Err( (vlc_object_t*) p_this, "D-Bus message reading : %s",
                error.message );
        dbus_error_free( &error );
        return DBUS_HANDLER_RESULT_NOT_YET_HANDLED;
    }

    p_playlist = pl_Hold( (vlc_object_t*) p_this );
    PL_LOCK;
    if( i_position < playlist_CurrentSize( p_playlist ) )
    {
        playlist_DeleteFromInput( p_playlist,
            p_playlist->current.p_elems[i_position]->p_input->i_id,
            pl_Locked );
    }
    PL_UNLOCK;

    pl_Release( (vlc_object_t*) p_this );

    REPLY_SEND;
}

DBUS_METHOD( SetLoop )
{
    REPLY_INIT;
    OUT_ARGUMENTS;

    DBusError error;
    dbus_bool_t b_loop;
    playlist_t* p_playlist;

    dbus_error_init( &error );
    dbus_message_get_args( p_from, &error,
            DBUS_TYPE_BOOLEAN, &b_loop,
            DBUS_TYPE_INVALID );

    if( dbus_error_is_set( &error ) )
    {
        msg_Err( (vlc_object_t*) p_this, "D-Bus message reading : %s",
                error.message );
        dbus_error_free( &error );
        return DBUS_HANDLER_RESULT_NOT_YET_HANDLED;
    }

    p_playlist = pl_Hold( (vlc_object_t*) p_this );
    var_SetBool( p_playlist, "loop", ( b_loop == TRUE ) );
    pl_Release( ((vlc_object_t*) p_this) );

    REPLY_SEND;
}

DBUS_METHOD( Repeat )
{
    REPLY_INIT;
    OUT_ARGUMENTS;

    DBusError error;
    dbus_bool_t b_repeat;
    playlist_t* p_playlist;

    dbus_error_init( &error );
    dbus_message_get_args( p_from, &error,
            DBUS_TYPE_BOOLEAN, &b_repeat,
            DBUS_TYPE_INVALID );

    if( dbus_error_is_set( &error ) )
    {
        msg_Err( (vlc_object_t*) p_this, "D-Bus message reading : %s",
                error.message );
        dbus_error_free( &error );
        return DBUS_HANDLER_RESULT_NOT_YET_HANDLED;
    }

    p_playlist = pl_Hold( (vlc_object_t*) p_this );
    var_SetBool( p_playlist, "repeat", ( b_repeat == TRUE ) );
    pl_Release( ((vlc_object_t*) p_this) );

    REPLY_SEND;
}

DBUS_METHOD( SetRandom )
{
    REPLY_INIT;
    OUT_ARGUMENTS;

    DBusError error;
    dbus_bool_t b_random;
    playlist_t* p_playlist;

    dbus_error_init( &error );
    dbus_message_get_args( p_from, &error,
            DBUS_TYPE_BOOLEAN, &b_random,
            DBUS_TYPE_INVALID );

    if( dbus_error_is_set( &error ) )
    {
        msg_Err( (vlc_object_t*) p_this, "D-Bus message reading : %s",
                error.message );
        dbus_error_free( &error );
        return DBUS_HANDLER_RESULT_NOT_YET_HANDLED;
    }

    p_playlist = pl_Hold( (vlc_object_t*) p_this );
    var_SetBool( p_playlist, "random", ( b_random == TRUE ) );
    pl_Release( ((vlc_object_t*) p_this) );

    REPLY_SEND;
}

DBUS_METHOD( GetRate )
{
    REPLY_INIT;
    OUT_ARGUMENTS;

    float f_rate = 0.0;

    playlist_t *p_playlist = pl_Hold( (vlc_object_t*) p_this );
    input_thread_t* p_input = playlist_CurrentInput( p_playlist );
    if( p_input )
    {
        int32_t i_rate;
        bool    b_can_rewind;

        i_rate = var_GetInteger( p_input, "rate" );
        b_can_rewind = var_GetBool( p_input, "can-rewind" );

        if( (i_rate < 0) && !b_can_rewind )
            f_rate = 0.0;
        else
            f_rate = (float)1000.0f/i_rate;

        vlc_object_release( p_input );
    }
    pl_Release( (vlc_object_t*) p_this );

    double rate = (double) f_rate;
    ADD_DOUBLE( &rate );
    REPLY_SEND;
}

DBUS_METHOD( SetRate )
{
    REPLY_INIT;
    OUT_ARGUMENTS;

    DBusError error;
    double rate = 0.0;
    dbus_int32_t i_success = -1;

    dbus_error_init( &error );
    dbus_message_get_args( p_from, &error,
            DBUS_TYPE_DOUBLE, &rate,
            DBUS_TYPE_INVALID );

    if( dbus_error_is_set( &error ) )
    {
        msg_Err( (vlc_object_t*) p_this, "D-Bus message reading : %s",
                error.message );
        dbus_error_free( &error );
        return DBUS_HANDLER_RESULT_NOT_YET_HANDLED;
    }

    playlist_t *p_playlist = pl_Hold( (vlc_object_t*) p_this );
    input_thread_t* p_input = playlist_CurrentInput( p_playlist );
    if( p_input )
    {
        bool    b_can_rewind;

        b_can_rewind = var_GetBool( p_input, "can-rewind" );
        if( (rate < 0.0) && !b_can_rewind )
            i_success = -1;
        else
        {
            int32_t i_rate = 1000.0f/(float)rate;
            int i_ret = var_SetInteger( p_input, "rate", i_rate );
            if( i_ret == 0 )
                 i_success = 0;
        }
        vlc_object_release( p_input );
    }
    pl_Release( ((vlc_object_t*) p_this) );

    ADD_INT32( &i_success );
    REPLY_SEND;
}

/*
 * /ES object method
 */

DBUS_METHOD( GetVideoES )
{
    REPLY_INIT;
    OUT_ARGUMENTS;
    dbus_int32_t i_es = 0;

    playlist_t *p_playlist = pl_Hold( (vlc_object_t*) p_this );
    input_thread_t *p_input = playlist_CurrentInput( p_playlist );
    if( p_input )
    {
        vlc_value_t val;
        int i_ret = -1;

        i_ret = var_Get( p_input, "video-es", &val );
        if( i_ret >= 0 )
        {
            vlc_value_t val_list;
            int i;

            var_Change( p_input, "video-es", VLC_VAR_GETCHOICES, &val_list, NULL );
            for( i = 0; i < val_list.p_list->i_count; i++ )
            {
                vlc_value_t es_val = val_list.p_list->p_values[i];
                if( es_val.i_int == val.i_int )
                {
                    i_es = i;
                    break;
                }
            }
        }
        vlc_object_release( p_input );
    }
    pl_Release( (vlc_object_t*) p_this );

    ADD_INT32( &i_es );
    REPLY_SEND;
}

DBUS_METHOD( SetVideoES )
{
    REPLY_INIT;
    OUT_ARGUMENTS;

    DBusError error;
    dbus_int32_t i_es = 0;
    dbus_int32_t i_success = -1;

    dbus_error_init( &error );
    dbus_message_get_args( p_from, &error,
            DBUS_TYPE_INT32, &i_es,
            DBUS_TYPE_INVALID );

    if( dbus_error_is_set( &error ) )
    {
        msg_Err( (vlc_object_t*) p_this, "D-Bus message reading : %s",
                error.message );
        dbus_error_free( &error );
        return DBUS_HANDLER_RESULT_NOT_YET_HANDLED;
    }

    playlist_t *p_playlist = pl_Hold( (vlc_object_t*) p_this );
    input_thread_t *p_input = playlist_CurrentInput( p_playlist );
    if( p_input )
    {
        vlc_value_t val_list;
        vlc_value_t newval;

        var_Change( p_input, "video-es", VLC_VAR_GETCHOICES, &val_list, NULL );
        if( (i_es < 0) || (i_es >= val_list.p_list->i_count) )
            i_es = -1;
        else
        {
            int i_ret = -1;
            newval = val_list.p_list->p_values[i_es];
            i_ret = var_Set( p_input, "video-es", newval );
            if( i_ret >= 0 )
                i_success = 0;
        }
        vlc_object_release( p_input );
    }
    pl_Release( (vlc_object_t*) p_this );

    ADD_INT32( &i_success );
    REPLY_SEND;
}

DBUS_METHOD( CountVideoES )
{
    REPLY_INIT;
    OUT_ARGUMENTS;

    dbus_int32_t i_count = 0;

    playlist_t *p_playlist = pl_Hold( (vlc_object_t*) p_this );
    input_thread_t *p_input = playlist_CurrentInput( p_playlist );
    if( p_input )
    {
        vlc_value_t val_list;
        var_Change( p_input, "video-es", VLC_VAR_GETCHOICES, &val_list, NULL );
        i_count = (val_list.p_list->i_count > 0) ? (val_list.p_list->i_count - 1): 0;
        vlc_object_release( p_input );
    }
    pl_Release( (vlc_object_t*) p_this );

    ADD_INT32( &i_count );
    REPLY_SEND;
}

DBUS_METHOD( GetAudioES )
{
    REPLY_INIT;
    OUT_ARGUMENTS;

    dbus_int32_t i_es = -1;

    playlist_t *p_playlist = pl_Hold( (vlc_object_t*) p_this );
    input_thread_t *p_input = playlist_CurrentInput( p_playlist );
    if( p_input )
    {
        vlc_value_t val;
        int i_ret = -1;

        i_ret = var_Get( p_input, "audio-es", &val );
        if( i_ret >= 0 )
        {
            vlc_value_t val_list;
            int i;

            var_Change( p_input, "audio-es", VLC_VAR_GETCHOICES, &val_list, NULL );
            for( i = 0; i < val_list.p_list->i_count; i++ )
            {
                vlc_value_t es_val = val_list.p_list->p_values[i];
                if( es_val.i_int == val.i_int )
                {
                    i_es = i;
                    break;
                }
            }
        }
        vlc_object_release( p_input );
    }
    pl_Release( (vlc_object_t*) p_this );

    ADD_INT32( &i_es );
    REPLY_SEND;
}

DBUS_METHOD( SetAudioES )
{
    REPLY_INIT;
    OUT_ARGUMENTS;

    DBusError error;
    dbus_int32_t i_es = 0;
    dbus_int32_t i_success = -1;

    dbus_error_init( &error );
    dbus_message_get_args( p_from, &error,
            DBUS_TYPE_INT32, &i_es,
            DBUS_TYPE_INVALID );

    if( dbus_error_is_set( &error ) )
    {
        msg_Err( (vlc_object_t*) p_this, "D-Bus message reading : %s",
                error.message );
        dbus_error_free( &error );
        return DBUS_HANDLER_RESULT_NOT_YET_HANDLED;
    }

    playlist_t *p_playlist = pl_Hold( (vlc_object_t*) p_this );
    input_thread_t *p_input = playlist_CurrentInput( p_playlist );
    if( p_input )
    {
        vlc_value_t val_list;
        vlc_value_t newval;

        var_Change( p_input, "audio-es", VLC_VAR_GETCHOICES, &val_list, NULL );
        if( (i_es < 0) || (i_es >= val_list.p_list->i_count) )
            i_es = -1;
        else
        {
            int i_ret = -1;
            newval = val_list.p_list->p_values[i_es];
            i_ret = var_Set( p_input, "audio-es", newval );
            if( i_ret >= 0 )
                i_success = 0;;
        }
        vlc_object_release( p_input );
    }
    pl_Release( (vlc_object_t*) p_this );

    ADD_INT32(&i_success);
    REPLY_SEND;
}

DBUS_METHOD( CountAudioES )
{
    REPLY_INIT;
    OUT_ARGUMENTS;

    dbus_int32_t i_count = 0;

    playlist_t *p_playlist = pl_Hold( (vlc_object_t*) p_this );
    input_thread_t *p_input = playlist_CurrentInput( p_playlist );
    if( p_input )
    {
        vlc_value_t val_list;
        var_Change( p_input, "audio-es", VLC_VAR_GETCHOICES, &val_list, NULL );
        i_count = (val_list.p_list->i_count > 0) ? (val_list.p_list->i_count - 1): 0;
        vlc_object_release( p_input );
    }
    pl_Release( (vlc_object_t*) p_this );

    ADD_INT32( &i_count );
    REPLY_SEND;
}

DBUS_METHOD( GetSubtitleES )
{
    REPLY_INIT;
    OUT_ARGUMENTS;

    dbus_int32_t i_es = 0;

    playlist_t *p_playlist = pl_Hold( (vlc_object_t*) p_this );
    input_thread_t *p_input = playlist_CurrentInput( p_playlist );
    if( p_input )
    {
        vlc_value_t val;
        int i_ret = -1;

        i_ret = var_Get( p_input, "spu-es", &val );
        if( i_ret >= 0 )
        {
            vlc_value_t val_list;
            int i;

            var_Change( p_input, "spu-es", VLC_VAR_GETCHOICES, &val_list, NULL );
            for( i = 0; i < val_list.p_list->i_count; i++ )
            {
                vlc_value_t spu_val = val_list.p_list->p_values[i];
                if( val.i_int == spu_val.i_int )
                {
                    i_es = i;
                    break;
                }
            }
        }
        vlc_object_release( p_input );
    }
    pl_Release( (vlc_object_t*) p_this );

    ADD_INT32( &i_es );
    REPLY_SEND;
}

DBUS_METHOD( SetSubtitleES )
{
    REPLY_INIT;
    OUT_ARGUMENTS;

    DBusError error;
    dbus_int32_t i_es = 0;
    dbus_int32_t i_success = -1;

    dbus_error_init( &error );
    dbus_message_get_args( p_from, &error,
            DBUS_TYPE_INT32, &i_es,
            DBUS_TYPE_INVALID );

    if( dbus_error_is_set( &error ) )
    {
        msg_Err( (vlc_object_t*) p_this, "D-Bus message reading : %s",
                error.message );
        dbus_error_free( &error );
        return DBUS_HANDLER_RESULT_NOT_YET_HANDLED;
    }

    playlist_t *p_playlist = pl_Hold( (vlc_object_t*) p_this );
    input_thread_t *p_input = playlist_CurrentInput( p_playlist );
    if( p_input )
    {
        vlc_value_t val_list;
        vlc_value_t newval;
        int i_ret = -1;

        var_Change( p_input, "spu-es", VLC_VAR_GETCHOICES, &val_list, NULL );
        if( val_list.p_list->i_count == 0 )
            i_es = -1;

        if( (i_es < 0) || (i_es >= val_list.p_list->i_count) )
            i_es = -1;

        newval = val_list.p_list->p_values[i_es];
        i_ret = var_Set( p_input, "spu-es", newval );
        if( i_ret < 0 )
            i_es = -1;

        vlc_object_release( p_input );
    }
    pl_Release( (vlc_object_t*) p_this );

    if( i_es >= 0 )
        i_success = 0;
    ADD_INT32(&i_success);
    REPLY_SEND;
}

DBUS_METHOD( CountSubtitleES )
{
    REPLY_INIT;
    OUT_ARGUMENTS;

    dbus_int32_t i_count = 0;

    playlist_t *p_playlist = pl_Hold( (vlc_object_t*) p_this );
    input_thread_t *p_input = playlist_CurrentInput( p_playlist );
    if( p_input )
    {
        vlc_value_t val_list;
        var_Change( p_input, "spu-es", VLC_VAR_GETCHOICES, &val_list, NULL );
        i_count = (val_list.p_list->i_count > 0) ? (val_list.p_list->i_count - 1): 0;
        vlc_object_release( p_input );
    }
    pl_Release( (vlc_object_t*) p_this );

    ADD_INT32( &i_count );
    REPLY_SEND;
}

DBUS_METHOD( GetTeletextPage )
{
    REPLY_INIT;
    OUT_ARGUMENTS;

    dbus_int32_t i_page = 0;

    playlist_t *p_playlist = pl_Hold( (vlc_object_t*) p_this );
    input_thread_t *p_input = playlist_CurrentInput( p_playlist );
    if( p_input )
    {
        vout_thread_t *p_vout;
        p_vout = input_GetVout( p_input );
        if( p_vout )
        {
            vlc_object_t *p_vbi;
            p_vbi = (vlc_object_t *) vlc_object_find_name( p_vout, "zvbi",
                                                           FIND_CHILD );
            if( p_vbi )
            {
                i_page = var_GetInteger( p_vbi, "vbi-page" );
                vlc_object_release( p_vbi );
            }
            vlc_object_release( p_vout );
        }
        vlc_object_release( p_input );
    }
    pl_Release( (vlc_object_t*) p_this );

    ADD_INT32( &i_page );
    REPLY_SEND;
}

DBUS_METHOD( SetTeletextPage )
{
    REPLY_INIT;
    OUT_ARGUMENTS;

    DBusError error;
    dbus_int32_t i_page = 0;
    dbus_int32_t i_success = -1;

    dbus_error_init( &error );
    dbus_message_get_args( p_from, &error,
            DBUS_TYPE_INT32, &i_page,
            DBUS_TYPE_INVALID );

    if( dbus_error_is_set( &error ) )
    {
        msg_Err( (vlc_object_t*) p_this, "D-Bus message reading : %s",
                error.message );
        dbus_error_free( &error );
        return DBUS_HANDLER_RESULT_NOT_YET_HANDLED;
    }

    playlist_t *p_playlist = pl_Hold( (vlc_object_t*) p_this );
    input_thread_t *p_input = playlist_CurrentInput( p_playlist );
    if( p_input )
    {
        vout_thread_t *p_vout;
        p_vout = input_GetVout( p_input );
        if( p_vout )
        {
            vlc_object_t *p_vbi;
            p_vbi = (vlc_object_t *) vlc_object_find_name( p_vout, "zvbi",
                                                           FIND_CHILD );
            if( p_vbi )
            {
                int i_ret;
                i_ret = var_SetInteger( p_vbi, "vbi-page", i_page );
                vlc_object_release( p_vbi );
                if( i_ret < 0 )
                    i_page = 0;
                else
                    i_success = 0;
            }
            vlc_object_release( p_vout );
        }
        vlc_object_release( p_input );
    }
    pl_Release( (vlc_object_t*) p_this );

    ADD_INT32(&i_success);
    REPLY_SEND;
}

DBUS_METHOD( TransparentTeletext )
{
    REPLY_INIT;
    OUT_ARGUMENTS;

    DBusError error;
    dbus_bool_t b_transparent = false;

    dbus_error_init( &error );
    dbus_message_get_args( p_from, &error,
            DBUS_TYPE_BOOLEAN, &b_transparent,
            DBUS_TYPE_INVALID );

    if( dbus_error_is_set( &error ) )
    {
        msg_Err( (vlc_object_t*) p_this, "D-Bus message reading : %s",
                error.message );
        dbus_error_free( &error );
        return DBUS_HANDLER_RESULT_NOT_YET_HANDLED;
    }

    playlist_t *p_playlist = pl_Hold( (vlc_object_t*) p_this );
    input_thread_t *p_input = playlist_CurrentInput( p_playlist );
    if( p_input )
    {
        vout_thread_t *p_vout;
        p_vout = input_GetVout( p_input );
        if( p_vout )
        {
            int i_ret = -1;
            if( var_CountChoices( p_input, "teletext-es" ) > 0 )
            {
                const bool b_selected = var_GetInteger( p_input, "teletext-es" ) >= 0;
                vlc_object_t *p_vbi;
                p_vbi = (vlc_object_t *)vlc_object_find_name( p_input, "zvbi",
                                                              FIND_CHILD );
                if( p_vbi )
                {
                    if( b_selected )
                    {
                        /* FIXME Gni, why that ? */
                        i_ret = var_SetInteger( p_vbi, "vbi-page",
                                                var_GetInteger( p_vbi, "vbi-page" ) );
                        if( i_ret < 0 )
                            b_transparent = FALSE;
                    }
                    else
                    {
                        /* FIXME Gni^2 */
                        i_ret = var_SetBool( p_vbi, "vbi-opaque",
                                            !var_GetBool( p_vbi, "vbi-opaque" ) );
                        if( i_ret < 0 )
                            b_transparent = FALSE;
                    }
                    vlc_object_release( p_vbi );
                }
                else if( b_selected )
                {
                    var_SetInteger( p_input, "spu-es", -1 );
                }
                else
                {
                    vlc_value_t list;
                    if( !var_Change( p_input, "teletext-es", VLC_VAR_GETLIST, &list, NULL ) )
                    {
                        if( list.p_list->i_count > 0 )
                            var_SetInteger( p_input, "spu-es", list.p_list->p_values[0].i_int );

                        var_Change( p_input, "teletext-es", VLC_VAR_FREELIST, &list, NULL );
                    }
                }
            }
            vlc_object_release( p_vout );
        }
        vlc_object_release( p_input );
    }
    pl_Release( (vlc_object_t*) p_this );

    REPLY_SEND;
}

DBUS_METHOD( GetEsMetadata )
{
    REPLY_INIT;
    OUT_ARGUMENTS;
    DBusError error;
    dbus_error_init( &error );

    dbus_int32_t i_es;

    dbus_message_get_args( p_from, &error,
           DBUS_TYPE_INT32, &i_es,
           DBUS_TYPE_INVALID );

    if( dbus_error_is_set( &error ) )
    {
        msg_Err( (vlc_object_t*) p_this, "D-Bus message reading : %s",
                error.message );
        dbus_error_free( &error );
        return DBUS_HANDLER_RESULT_NOT_YET_HANDLED;
    }

    playlist_t *p_playlist = pl_Hold( (vlc_object_t*) p_this );
    PL_LOCK;
    playlist_item_t* p_item = playlist_CurrentPlayingItem( p_playlist );
    if( p_item )
        GetInputEsMeta( p_item->p_input, i_es, &args );
    PL_UNLOCK;
    pl_Release( (vlc_object_t*) p_this );

    REPLY_SEND;
}

/*
 * /Display object method
 */

DBUS_METHOD( GetVideoPosition )
{
    REPLY_INIT;
    OUT_ARGUMENTS;

    dbus_int32_t i_x = 0;
    dbus_int32_t i_y = 0;
    dbus_int32_t i_width = 0;
    dbus_int32_t i_height = 0;

    playlist_t *p_playlist = pl_Hold( (vlc_object_t*) p_this );
    input_thread_t *p_input = playlist_CurrentInput( p_playlist );
    if( p_input )
    {
        vout_thread_t *p_vout;
        p_vout = input_GetVout( p_input );
        if( p_vout )
        {
            i_x = var_GetInteger( p_vout, "video-x");
            i_y = var_GetInteger( p_vout, "video-y" );
            i_width = var_GetInteger( p_vout, "width" );
            if (i_width == -1) i_width = p_vout->i_window_width;
            i_height = var_GetInteger( p_vout, "height" );
            if (i_height == -1) i_height = p_vout->i_window_height;
            vlc_object_release( p_vout );
        }
        vlc_object_release( p_input );
    }
    pl_Release( (vlc_object_t*) p_this );

    ADD_INT32( &i_x );
    ADD_INT32( &i_y );
    ADD_INT32( &i_width );
    ADD_INT32( &i_height );

    REPLY_SEND;
}

DBUS_METHOD( SetVideoPosition )
{
    REPLY_INIT;
    OUT_ARGUMENTS;

    DBusError error;
    dbus_int32_t i_x = 0;
    dbus_int32_t i_y = 0;
    dbus_int32_t i_width = 0;
    dbus_int32_t i_height = 0;
    dbus_int32_t i_success = -1;

    dbus_error_init( &error );
    dbus_message_get_args( p_from, &error,
            DBUS_TYPE_INT32, &i_x,
            DBUS_TYPE_INT32, &i_y,
            DBUS_TYPE_INT32, &i_width,
            DBUS_TYPE_INT32, &i_height,
            DBUS_TYPE_INVALID );

    if( dbus_error_is_set( &error ) )
    {
        msg_Err( (vlc_object_t*) p_this, "D-Bus message reading : %s",
                error.message );
        dbus_error_free( &error );
        return DBUS_HANDLER_RESULT_NOT_YET_HANDLED;
    }

    playlist_t *p_playlist = pl_Hold( (vlc_object_t*) p_this );
    input_thread_t *p_input = playlist_CurrentInput( p_playlist );
    if( p_input )
    {
        vout_thread_t *p_vout;
        p_vout = input_GetVout( p_input );
        if( p_vout )
        {
            int i_ret = 0;
            /* NOTE: this doesn't work when the video output is embedded
             * in the main window. Use: vlc --no-embedded
             */
            var_SetInteger( p_vout, "video-x", i_x );
            var_SetInteger( p_vout, "video-y", i_y );
            i_ret = vout_Control( p_vout, VOUT_SET_POSITION, i_x, i_y);
            if( i_ret >= 0 )
            {
                var_SetInteger( p_vout, "width", i_width );
                var_SetInteger( p_vout, "height", i_height );

                i_ret = vout_Control( p_vout, VOUT_SET_SIZE, i_width, i_height);
                if( i_ret >= 0 )
                {
                    i_success = 0;
                }
            }
            vlc_object_release( p_vout );
        }
        vlc_object_release( p_input );
    }
    pl_Release( (vlc_object_t*) p_this );

    ADD_INT32(&i_success);
    REPLY_SEND;
}

DBUS_METHOD( GetAspectRatio )
{
    REPLY_INIT;
    OUT_ARGUMENTS;

    char *psz_aspect = NULL;

    playlist_t *p_playlist = pl_Hold( (vlc_object_t*) p_this );
    input_thread_t *p_input = playlist_CurrentInput( p_playlist );
    if( p_input )
    {
        vout_thread_t *p_vout;
        p_vout = input_GetVout( p_input );
        if( p_vout )
        {
            psz_aspect = var_GetNonEmptyString( p_vout, "aspect-ratio" );
            vlc_object_release( p_vout );
        }
        vlc_object_release( p_input );
    }
    pl_Release( (vlc_object_t*) p_this );

    if( !psz_aspect )
        psz_aspect = strdup( "default" );

    ADD_STRING( &psz_aspect );
    free( psz_aspect );

    REPLY_SEND;
}

DBUS_METHOD( SetAspectRatio )
{
    REPLY_INIT;
    OUT_ARGUMENTS;

    DBusError error;
    char *psz_aspect = NULL;
    dbus_int32_t i_success = -1;

    dbus_error_init( &error );
    dbus_message_get_args( p_from, &error,
            DBUS_TYPE_STRING, &psz_aspect,
            DBUS_TYPE_INVALID );

    if( dbus_error_is_set( &error ) )
    {
        msg_Err( (vlc_object_t*) p_this, "D-Bus message reading : %s",
                error.message );
        dbus_error_free( &error );
        return DBUS_HANDLER_RESULT_NOT_YET_HANDLED;
    }

    playlist_t *p_playlist = pl_Hold( (vlc_object_t*) p_this );
    input_thread_t *p_input = playlist_CurrentInput( p_playlist );
    if( p_input )
    {
        vout_thread_t *p_vout;
        p_vout = input_GetVout( p_input );
        if( p_vout )
        {
            if( psz_aspect )
            {
                int i_ret = var_SetString( p_vout, "aspect-ratio", psz_aspect );
                if( i_ret >= 0 )
                    i_success = 0;
            }
            vlc_object_release( p_vout );
        }
        vlc_object_release( p_input );
    }
    pl_Release( (vlc_object_t*) p_this );

    ADD_INT32(&i_success);
    REPLY_SEND;
}

DBUS_METHOD( GetAudioChannel )
{
    REPLY_INIT;
    OUT_ARGUMENTS;

    dbus_int32_t i_channel;

    playlist_t *p_playlist = pl_Hold( (vlc_object_t*) p_this );
    input_thread_t *p_input = playlist_CurrentInput( p_playlist );
    if( p_input )
    {
        aout_instance_t *p_aout;
        p_aout = input_GetAout( p_input );
        if( p_aout )
        {
            i_channel = var_GetInteger( p_aout, "audio-channels" );
            vlc_object_release( p_aout );
        }
        vlc_object_release( p_input );
    }
    pl_Release( (vlc_object_t*) p_this );

    ADD_INT32( &i_channel );

    REPLY_SEND;
}

DBUS_METHOD( SetAudioChannel )
{
    REPLY_INIT;
    OUT_ARGUMENTS;

    DBusError error;
    dbus_int32_t i_channel;
    dbus_int32_t i_success = -1;

    dbus_error_init( &error );
    dbus_message_get_args( p_from, &error,
            DBUS_TYPE_INT32, &i_channel,
            DBUS_TYPE_INVALID );

    if( dbus_error_is_set( &error ) )
    {
        msg_Err( (vlc_object_t*) p_this, "D-Bus message reading : %s",
                error.message );
        dbus_error_free( &error );
        return DBUS_HANDLER_RESULT_NOT_YET_HANDLED;
    }

    playlist_t *p_playlist = pl_Hold( (vlc_object_t*) p_this );
    input_thread_t *p_input = playlist_CurrentInput( p_playlist );
    if( p_input )
    {
        aout_instance_t *p_aout;
        p_aout = input_GetAout( p_input );
        if( p_aout )
        {
            int i_ret = var_SetInteger( p_aout, "audo-channels", i_channel );
            if( i_ret >= 0 )
                i_success = 0;
            vlc_object_release( p_aout );
        }
        vlc_object_release( p_input );
    }
    pl_Release( (vlc_object_t*) p_this );

    ADD_INT32(&i_success);
    REPLY_SEND;
}

DBUS_METHOD( Fullscreen )
{
    REPLY_INIT;

    playlist_t *p_playlist = pl_Hold( (vlc_object_t*) p_this );
    input_thread_t *p_input =  playlist_CurrentInput( p_playlist );
    if( p_input )
    {
        vout_thread_t *p_vout;
        p_vout = input_GetVout( p_input );
        if( p_vout )
        {
            bool b_fullscreen = var_GetBool( p_vout, "fullscreen" );
            var_SetBool( p_vout, "fullscreen", !b_fullscreen );
            vlc_object_release( p_vout );
        }
        vlc_object_release( p_input );
    }
    pl_Release( (vlc_object_t*) p_this );

    REPLY_SEND;
}

/*****************************************************************************
 * Introspection method
 *****************************************************************************/

DBUS_METHOD( handle_introspect_root )
{ /* handles introspection of root object */
    VLC_UNUSED(p_this);
    REPLY_INIT;
    OUT_ARGUMENTS;
    ADD_STRING( &psz_introspection_xml_data_root );
    REPLY_SEND;
}

DBUS_METHOD( handle_introspect_player )
{
    VLC_UNUSED(p_this);
    REPLY_INIT;
    OUT_ARGUMENTS;
    ADD_STRING( &psz_introspection_xml_data_player );
    REPLY_SEND;
}

DBUS_METHOD( handle_introspect_tracklist )
{
    VLC_UNUSED(p_this);
    REPLY_INIT;
    OUT_ARGUMENTS;
    ADD_STRING( &psz_introspection_xml_data_tracklist );
    REPLY_SEND;
}

DBUS_METHOD( handle_introspect_es )
{
    VLC_UNUSED(p_this);
    REPLY_INIT;
    OUT_ARGUMENTS;
    ADD_STRING( &psz_introspection_xml_data_es );
    REPLY_SEND;
}

DBUS_METHOD( handle_introspect_display )
{
    VLC_UNUSED(p_this);
    REPLY_INIT;
    OUT_ARGUMENTS;
    ADD_STRING( &psz_introspection_xml_data_display );
    REPLY_SEND;
}

/*****************************************************************************
 * handle_*: answer to incoming messages
 *****************************************************************************/

#define METHOD_FUNC( method, function ) \
    else if( dbus_message_is_method_call( p_from, MPRIS_DBUS_INTERFACE, method ) )\
        return function( p_conn, p_from, p_this )

DBUS_METHOD( handle_root )
{

    if( dbus_message_is_method_call( p_from,
                DBUS_INTERFACE_INTROSPECTABLE, "Introspect" ) )
        return handle_introspect_root( p_conn, p_from, p_this );

    /* here D-Bus method's names are associated to an handler */

    METHOD_FUNC( "Identity",                Identity );
    METHOD_FUNC( "MprisVersion",            MprisVersion );
    METHOD_FUNC( "Quit",                    Quit );

    return DBUS_HANDLER_RESULT_NOT_YET_HANDLED;
}

DBUS_METHOD( handle_player )
{
    if( dbus_message_is_method_call( p_from,
                DBUS_INTERFACE_INTROSPECTABLE, "Introspect" ) )
        return handle_introspect_player( p_conn, p_from, p_this );

    /* here D-Bus method's names are associated to an handler */

    METHOD_FUNC( "Prev",                    Prev );
    METHOD_FUNC( "Next",                    Next );
    METHOD_FUNC( "Stop",                    Stop );
    METHOD_FUNC( "Play",                    Play );
    METHOD_FUNC( "Pause",                   Pause );
    METHOD_FUNC( "Repeat",                  Repeat );
    METHOD_FUNC( "VolumeSet",               VolumeSet );
    METHOD_FUNC( "VolumeGet",               VolumeGet );
    METHOD_FUNC( "PositionSet",             PositionSet );
    METHOD_FUNC( "PositionGet",             PositionGet );
    METHOD_FUNC( "GetStatus",               GetStatus );
    METHOD_FUNC( "GetMetadata",             GetCurrentMetadata );
    METHOD_FUNC( "GetCaps",                 GetCaps );
    METHOD_FUNC( "GetRate",                 GetRate );
    METHOD_FUNC( "SetRate",                 SetRate );

    return DBUS_HANDLER_RESULT_NOT_YET_HANDLED;
}

DBUS_METHOD( handle_tracklist )
{
    if( dbus_message_is_method_call( p_from,
                DBUS_INTERFACE_INTROSPECTABLE, "Introspect" ) )
    return handle_introspect_tracklist( p_conn, p_from, p_this );

    /* here D-Bus method's names are associated to an handler */

    METHOD_FUNC( "GetMetadata",             GetMetadata );
    METHOD_FUNC( "GetCurrentTrack",         GetCurrentTrack );
    METHOD_FUNC( "GetLength",               GetLength );
    METHOD_FUNC( "AddTrack",                AddTrack );
    METHOD_FUNC( "DelTrack",                DelTrack );
    METHOD_FUNC( "SetLoop",                 SetLoop );
    METHOD_FUNC( "SetRandom",               SetRandom );

    return DBUS_HANDLER_RESULT_NOT_YET_HANDLED;
}

DBUS_METHOD( handle_es )
{
    if( dbus_message_is_method_call( p_from,
                DBUS_INTERFACE_INTROSPECTABLE, "Introspect" ) )
    return handle_introspect_es( p_conn, p_from, p_this );

    /* here D-Bus method's names are associated to an handler */

    METHOD_FUNC( "GetVideoES",              GetVideoES );
    METHOD_FUNC( "SetVideoES",              SetVideoES );
    METHOD_FUNC( "CountVideoES",            CountVideoES );

    METHOD_FUNC( "GetAudioES",              GetAudioES );
    METHOD_FUNC( "SetAudioES",              SetAudioES );
    METHOD_FUNC( "CountAudioES",            CountAudioES );

    METHOD_FUNC( "GetSubtitleES",           GetSubtitleES );
    METHOD_FUNC( "SetSubtitleES",           SetSubtitleES );
    METHOD_FUNC( "CountSubtitleES",         CountSubtitleES );

    METHOD_FUNC( "GetTeletextPage",         GetTeletextPage );
    METHOD_FUNC( "SetTeletextPage",         SetTeletextPage );
    METHOD_FUNC( "TransparentTeletext",     TransparentTeletext );

    METHOD_FUNC( "GetEsMetadata",           GetEsMetadata );

    return DBUS_HANDLER_RESULT_NOT_YET_HANDLED;
}

DBUS_METHOD( handle_display )
{
    if( dbus_message_is_method_call( p_from,
                DBUS_INTERFACE_INTROSPECTABLE, "Introspect" ) )
    return handle_introspect_display( p_conn, p_from, p_this );

    /* here D-Bus method's names are associated to an handler */

    METHOD_FUNC( "GetVideoPosition",        GetVideoPosition );
    METHOD_FUNC( "SetVideoPosition",        SetVideoPosition );
    METHOD_FUNC( "GetAspectRatio",          GetAspectRatio );
    METHOD_FUNC( "SetAspectRatio",          SetAspectRatio );
    METHOD_FUNC( "GetAudioChannel",         GetAudioChannel );
    METHOD_FUNC( "SetAudioChannel",         SetAudioChannel );
    METHOD_FUNC( "Fullscreen",              Fullscreen );

    return DBUS_HANDLER_RESULT_NOT_YET_HANDLED;
}

/*****************************************************************************
 * Open: initialize interface
 *****************************************************************************/

static int Open( vlc_object_t *p_this )
{ /* initialisation of the connection */
    intf_thread_t   *p_intf = (intf_thread_t*)p_this;
    intf_sys_t      *p_sys  = malloc( sizeof( intf_sys_t ) );
    playlist_t      *p_playlist;
    DBusConnection  *p_conn;
    DBusError       error;

    if( !p_sys )
        return VLC_ENOMEM;

    p_sys->b_meta_read = false;
    p_sys->i_caps = CAPS_NONE;
    p_sys->b_dead = false;

    dbus_error_init( &error );

    /* connect to the session bus */
    p_conn = dbus_bus_get( DBUS_BUS_SESSION, &error );
    if( !p_conn )
    {
        msg_Err( p_this, "Failed to connect to the D-Bus session daemon: %s",
                error.message );
        dbus_error_free( &error );
        free( p_sys );
        return VLC_EGENERIC;
    }

    /* register a well-known name on the bus */
    dbus_bus_request_name( p_conn, VLC_MPRIS_DBUS_SERVICE, 0, &error );
    if( dbus_error_is_set( &error ) )
    {
        msg_Err( p_this, "Error requesting service " VLC_MPRIS_DBUS_SERVICE
                 ": %s", error.message );
        dbus_error_free( &error );
        free( p_sys );
        return VLC_EGENERIC;
    }

    /* we register the objects */
    dbus_connection_register_object_path( p_conn, MPRIS_DBUS_ROOT_PATH,
            &vlc_dbus_root_vtable, p_this );
    dbus_connection_register_object_path( p_conn, MPRIS_DBUS_PLAYER_PATH,
            &vlc_dbus_player_vtable, p_this );
    dbus_connection_register_object_path( p_conn, MPRIS_DBUS_TRACKLIST_PATH,
            &vlc_dbus_tracklist_vtable, p_this );
    dbus_connection_register_object_path( p_conn, MPRIS_DBUS_ES_PATH,
            &vlc_dbus_es_vtable, p_this );
    dbus_connection_register_object_path( p_conn, MPRIS_DBUS_DISPLAY_PATH,
            &vlc_dbus_display_vtable, p_this );

    dbus_connection_flush( p_conn );

    p_intf->pf_run = Run;
    p_intf->p_sys = p_sys;
    p_sys->p_conn = p_conn;
    p_sys->p_events = vlc_array_new();
    vlc_mutex_init( &p_sys->lock );

    p_playlist = pl_Hold( p_intf );
    PL_LOCK;
    var_AddCallback( p_playlist, "item-current", AllCallback, p_intf );
    var_AddCallback( p_playlist, "intf-change", AllCallback, p_intf );
    var_AddCallback( p_playlist, "playlist-item-append", AllCallback, p_intf );
    var_AddCallback( p_playlist, "playlist-item-deleted", AllCallback, p_intf );
    var_AddCallback( p_playlist, "random", AllCallback, p_intf );
    var_AddCallback( p_playlist, "repeat", AllCallback, p_intf );
    var_AddCallback( p_playlist, "loop", AllCallback, p_intf );
    PL_UNLOCK;
    pl_Release( p_intf );

    UpdateCaps( p_intf );

    return VLC_SUCCESS;
}

/*****************************************************************************
 * Close: destroy interface
 *****************************************************************************/

static void Close   ( vlc_object_t *p_this )
{
    intf_thread_t   *p_intf     = (intf_thread_t*) p_this;
    playlist_t      *p_playlist = pl_Hold( p_intf );;
    input_thread_t  *p_input;

    var_DelCallback( p_playlist, "item-current", AllCallback, p_intf );
    var_DelCallback( p_playlist, "intf-change", AllCallback, p_intf );
    var_DelCallback( p_playlist, "playlist-item-append", AllCallback, p_intf );
    var_DelCallback( p_playlist, "playlist-item-deleted", AllCallback, p_intf );
    var_DelCallback( p_playlist, "random", AllCallback, p_intf );
    var_DelCallback( p_playlist, "repeat", AllCallback, p_intf );
    var_DelCallback( p_playlist, "loop", AllCallback, p_intf );

    p_input = playlist_CurrentInput( p_playlist );
    if ( p_input )
    {
        var_DelCallback( p_input, "state", AllCallback, p_intf );
        vlc_object_release( p_input );
    }

    pl_Release( p_intf );

    dbus_connection_unref( p_intf->p_sys->p_conn );

    // Free the events array
    for( int i = 0; i < vlc_array_count( p_intf->p_sys->p_events ); i++ )
    {
        callback_info_t* info = vlc_array_item_at_index( p_intf->p_sys->p_events, i );
        free( info );
    }
    vlc_mutex_destroy( &p_intf->p_sys->lock );
    vlc_array_destroy( p_intf->p_sys->p_events );
    free( p_intf->p_sys );
}

/*****************************************************************************
 * Run: main loop
 *****************************************************************************/

static void Run          ( intf_thread_t *p_intf )
{
    for( ;; )
    {
        if( dbus_connection_get_dispatch_status(p_intf->p_sys->p_conn)
                                             == DBUS_DISPATCH_COMPLETE )
            msleep( INTF_IDLE_SLEEP );
        int canc = vlc_savecancel();
        dbus_connection_read_write_dispatch( p_intf->p_sys->p_conn, 0 );

        // Get the messages
        vlc_mutex_lock( &p_intf->p_sys->lock );
        for( int i = vlc_array_count( p_intf->p_sys->p_events ) - 1; i >= 0; i-- )
        {
            callback_info_t* info = vlc_array_item_at_index( p_intf->p_sys->p_events, i );
            switch( info->signal )
            {
            case SIGNAL_ITEM_CURRENT:
                TrackChange( p_intf );
                break;
            case SIGNAL_INTF_CHANGE:
            case SIGNAL_PLAYLIST_ITEM_APPEND:
            case SIGNAL_PLAYLIST_ITEM_DELETED:
                TrackListChangeEmit( p_intf, info->signal, info->i_node );
                break;
            case SIGNAL_RANDOM:
            case SIGNAL_REPEAT:
            case SIGNAL_LOOP:
                StatusChangeEmit( p_intf );
                break;
            case SIGNAL_STATE:
                StateChange( p_intf, info->i_input_state );
                break;
            default:
                assert(0);
            }
            free( info );
            vlc_array_remove( p_intf->p_sys->p_events, i );
        }
        vlc_mutex_unlock( &p_intf->p_sys->lock );
        vlc_restorecancel( canc );
    }
}


// Get all the callbacks
static int AllCallback( vlc_object_t *p_this, const char *psz_var,
                        vlc_value_t oldval, vlc_value_t newval, void *p_data )
{
    (void)p_this;
    (void)oldval;
    intf_thread_t *p_intf = (intf_thread_t*)p_data;

    callback_info_t *info = malloc( sizeof( callback_info_t ) );
    if( !info )
        return VLC_ENOMEM;

    // Wich event is it ?
    if( !strcmp( "item-current", psz_var ) )
        info->signal = SIGNAL_ITEM_CURRENT;
    else if( !strcmp( "intf-change", psz_var ) )
        info->signal = SIGNAL_INTF_CHANGE;
    else if( !strcmp( "playlist-item-append", psz_var ) )
    {
        info->signal = SIGNAL_PLAYLIST_ITEM_APPEND;
        info->i_node = ((playlist_add_t*)newval.p_address)->i_node;
    }
    else if( !strcmp( "playlist-item-deleted", psz_var ) )
        info->signal = SIGNAL_PLAYLIST_ITEM_DELETED;
    else if( !strcmp( "random", psz_var ) )
        info->signal = SIGNAL_RANDOM;
    else if( !strcmp( "repeat", psz_var ) )
        info->signal = SIGNAL_REPEAT;
    else if( !strcmp( "loop", psz_var ) )
        info->signal = SIGNAL_LOOP;
    else if( !strcmp( "state", psz_var ) )
    {
        info->signal = SIGNAL_STATE;
        info->i_input_state = newval.i_int;
    }
    else
        assert(0);

    // Append the event
    vlc_mutex_lock( &p_intf->p_sys->lock );
    vlc_array_append( p_intf->p_sys->p_events, info );
    vlc_mutex_unlock( &p_intf->p_sys->lock );
    return VLC_SUCCESS;
}

/******************************************************************************
 * CapsChange: player capabilities change signal
 *****************************************************************************/
DBUS_SIGNAL( CapsChangeSignal )
{
    SIGNAL_INIT( "CapsChange" );
    OUT_ARGUMENTS;

    ADD_INT32( &((intf_thread_t*)p_data)->p_sys->i_caps );
    SIGNAL_SEND;
}

/******************************************************************************
 * TrackListChange: tracklist order / length change signal 
 *****************************************************************************/
DBUS_SIGNAL( TrackListChangeSignal )
{ /* emit the new tracklist lengh */
    SIGNAL_INIT("TrackListChange");
    OUT_ARGUMENTS;

    playlist_t *p_playlist = pl_Hold( (vlc_object_t*) p_data );
    PL_LOCK;
    dbus_int32_t i_elements = playlist_CurrentSize( p_playlist );
    PL_UNLOCK;
    pl_Release( (vlc_object_t*) p_data );

    ADD_INT32( &i_elements );
    SIGNAL_SEND;
}

/*****************************************************************************
 * TrackListChangeEmit: Emits the TrackListChange signal
 *****************************************************************************/
/* FIXME: It is not called on tracklist reordering */
static int TrackListChangeEmit( intf_thread_t *p_intf, int signal, int i_node )
{
    // "playlist-item-append"
    if( signal == SIGNAL_PLAYLIST_ITEM_APPEND )
    {
        /* don't signal when items are added/removed in p_category */
        playlist_t *p_playlist = pl_Hold( p_intf );
        PL_LOCK;
        playlist_item_t *p_item = playlist_ItemGetById( p_playlist, i_node );
        assert( p_item );
        while( p_item->p_parent )
            p_item = p_item->p_parent;
        if( p_item == p_playlist->p_root_category )
        {
            PL_UNLOCK;
            pl_Release( p_intf );
            return VLC_SUCCESS;
        }
        PL_UNLOCK;
        pl_Release( p_intf );
    }

    if( p_intf->p_sys->b_dead )
        return VLC_SUCCESS;

    UpdateCaps( p_intf );
    TrackListChangeSignal( p_intf->p_sys->p_conn, p_intf );
    return VLC_SUCCESS;
}
/*****************************************************************************
 * TrackChange: Playlist item change callback
 *****************************************************************************/

DBUS_SIGNAL( TrackChangeSignal )
{ /* emit the metadata of the new item */
    SIGNAL_INIT( "TrackChange" );
    OUT_ARGUMENTS;

    input_item_t *p_item = (input_item_t*) p_data;
    GetInputMeta ( p_item, &args );

    SIGNAL_SEND;
}

/*****************************************************************************
 * StatusChange: Player status change signal
 *****************************************************************************/

DBUS_SIGNAL( StatusChangeSignal )
{ /* send the updated status info on the bus */
    SIGNAL_INIT( "StatusChange" );
    OUT_ARGUMENTS;

    /* we're called from a callback of input_thread_t, so it can not be
     * destroyed before we return */
    MarshalStatus( (intf_thread_t*) p_data, &args, false );

    SIGNAL_SEND;
}

/*****************************************************************************
 * StateChange: callback on input "state"
 *****************************************************************************/
//static int StateChange( vlc_object_t *p_this, const char* psz_var,
//            vlc_value_t oldval, vlc_value_t newval, void *p_data )
static int StateChange( intf_thread_t *p_intf, int i_input_state )
{
    intf_sys_t          *p_sys      = p_intf->p_sys;
    playlist_t          *p_playlist;
    input_thread_t      *p_input;
    input_item_t        *p_item;

    if( p_intf->p_sys->b_dead )
        return VLC_SUCCESS;

    UpdateCaps( p_intf );

    if( !p_sys->b_meta_read && i_input_state == PLAYING_S )
    {
        p_playlist = pl_Hold( p_intf );
        p_input = playlist_CurrentInput( p_playlist );
        if( p_input )
        {
            p_item = input_GetItem( p_input );
            if( p_item )
            {
                p_sys->b_meta_read = true;
                TrackChangeSignal( p_sys->p_conn, p_item );
            }
            vlc_object_release( p_input );
        }
        pl_Release( p_intf );
    }

    if( i_input_state == PLAYING_S || i_input_state == PAUSE_S ||
        i_input_state == END_S )
    {
        StatusChangeSignal( p_sys->p_conn, p_intf );
    }

    return VLC_SUCCESS;
}

/*****************************************************************************
 * StatusChangeEmit: Emits the StatusChange signal
 *****************************************************************************/
static int StatusChangeEmit( intf_thread_t * p_intf )
{
    if( p_intf->p_sys->b_dead )
        return VLC_SUCCESS;

    UpdateCaps( p_intf );
    StatusChangeSignal( p_intf->p_sys->p_conn, p_intf );
    return VLC_SUCCESS;
}

/*****************************************************************************
 * TrackChange: callback on playlist "item-current"
 *****************************************************************************/
static int TrackChange( intf_thread_t *p_intf )
{
    intf_sys_t          *p_sys      = p_intf->p_sys;
    playlist_t          *p_playlist;
    input_thread_t      *p_input    = NULL;
    input_item_t        *p_item     = NULL;

    if( p_intf->p_sys->b_dead )
        return VLC_SUCCESS;

    p_sys->b_meta_read = false;

    p_playlist = pl_Hold( p_intf );
    p_input = playlist_CurrentInput( p_playlist );
    if( !p_input )
    {
        pl_Release( p_intf );
        return VLC_SUCCESS;
    }

    pl_Release( p_intf );

    p_item = input_GetItem( p_input );
    if( !p_item )
    {
        vlc_object_release( p_input );
        return VLC_EGENERIC;
    }

    if( input_item_IsPreparsed( p_item ) )
    {
        p_sys->b_meta_read = true;
        TrackChangeSignal( p_sys->p_conn, p_item );
    }

    var_AddCallback( p_input, "state", AllCallback, p_intf );

    vlc_object_release( p_input );
    return VLC_SUCCESS;
}

/*****************************************************************************
 * UpdateCaps: update p_sys->i_caps
 * This function have to be called with the playlist unlocked
 ****************************************************************************/
static int UpdateCaps( intf_thread_t* p_intf )
{
    intf_sys_t* p_sys = p_intf->p_sys;
    dbus_int32_t i_caps = CAPS_CAN_HAS_TRACKLIST;
    playlist_t* p_playlist = pl_Hold( p_intf );

    PL_LOCK;
    if( p_playlist->current.i_size > 0 )
        i_caps |= CAPS_CAN_PLAY | CAPS_CAN_GO_PREV | CAPS_CAN_GO_NEXT;
    PL_UNLOCK;

    input_thread_t* p_input = playlist_CurrentInput( p_playlist );
    if( p_input )
    {
        /* XXX: if UpdateCaps() is called too early, these are
         * unconditionnaly true */
        if( var_GetBool( p_input, "can-pause" ) )
            i_caps |= CAPS_CAN_PAUSE;
        if( var_GetBool( p_input, "can-seek" ) )
            i_caps |= CAPS_CAN_SEEK;
        if( var_GetBool( p_input, "can-rate" ) )
            i_caps |= CAPS_CAN_CONTROL_RATE;
        if( var_GetBool( p_input, "can-rewind" ) )
            i_caps |= CAPS_CAN_CONTROL_REWIND;
        if( var_CountChoices( p_input, "spu-es" ) > 0 )
            i_caps |= CAPS_CAN_CONTROL_SUBTITLES;
        if( var_CountChoices( p_input, "teletext-es" ) > 0 )
            i_caps |= CAPS_CAN_CONTROL_TELETEXT;
        vlc_object_release( p_input );
    }

    pl_Release( p_intf );

    if( p_sys->b_meta_read )
        i_caps |= CAPS_CAN_PROVIDE_METADATA;

    if( i_caps != p_intf->p_sys->i_caps )
    {
        p_sys->i_caps = i_caps;
        CapsChangeSignal( p_intf->p_sys->p_conn, (vlc_object_t*)p_intf );
    }

    return VLC_SUCCESS;
}

/*****************************************************************************
 * GetInputMeta: Fill a DBusMessage with the given input item metadata
 *****************************************************************************/

#define ADD_META( entry, type, data ) \
    if( data ) { \
        dbus_message_iter_open_container( &dict, DBUS_TYPE_DICT_ENTRY, \
                NULL, &dict_entry ); \
        dbus_message_iter_append_basic( &dict_entry, DBUS_TYPE_STRING, \
                &ppsz_meta_items[entry] ); \
        dbus_message_iter_open_container( &dict_entry, DBUS_TYPE_VARIANT, \
                type##_AS_STRING, &variant ); \
        dbus_message_iter_append_basic( &variant, \
                type, \
                & data ); \
        dbus_message_iter_close_container( &dict_entry, &variant ); \
        dbus_message_iter_close_container( &dict, &dict_entry ); }

#define ADD_VLC_META_STRING( entry, item ) \
    { \
        char * psz = input_item_Get##item( p_input );\
        ADD_META( entry, DBUS_TYPE_STRING, \
                  psz ); \
        free( psz ); \
    }

static int GetInputMeta( input_item_t* p_input,
                        DBusMessageIter *args )
{
    DBusMessageIter dict, dict_entry, variant;
    /* We need the track length to be expressed in milli-seconds
     * instead of µ-seconds */
    dbus_int64_t i_length = ( input_item_GetDuration( p_input ) / 1000 );

    const char* ppsz_meta_items[] =
    {
    "title", "artist", "genre", "copyright", "album", "tracknum",
    "description", "rating", "date", "setting", "url", "language",
    "nowplaying", "publisher", "encodedby", "arturl", "trackid",
    "status", "location", "length", "video-codec", "audio-codec",
    "video-bitrate", "audio-bitrate", "audio-samplerate", "spu-codec"
    };

    dbus_message_iter_open_container( args, DBUS_TYPE_ARRAY, "{sv}", &dict );

    ADD_VLC_META_STRING( 0,  Title );
    ADD_VLC_META_STRING( 1,  Artist );
    ADD_VLC_META_STRING( 2,  Genre );
    ADD_VLC_META_STRING( 3,  Copyright );
    ADD_VLC_META_STRING( 4,  Album );
    ADD_VLC_META_STRING( 5,  TrackNum );
    ADD_VLC_META_STRING( 6,  Description );
    ADD_VLC_META_STRING( 7,  Rating );
    ADD_VLC_META_STRING( 8,  Date );
    ADD_VLC_META_STRING( 9,  Setting );
    ADD_VLC_META_STRING( 10, URL );
    ADD_VLC_META_STRING( 11, Language );
    ADD_VLC_META_STRING( 12, NowPlaying );
    ADD_VLC_META_STRING( 13, Publisher );
    ADD_VLC_META_STRING( 14, EncodedBy );
    ADD_VLC_META_STRING( 15, ArtURL );
    ADD_VLC_META_STRING( 16, TrackID );

    vlc_mutex_lock( &p_input->lock );
    if( p_input->p_meta )
        ADD_META( 17, DBUS_TYPE_INT32, p_input->p_meta->i_status );
    vlc_mutex_unlock( &p_input->lock );

    ADD_VLC_META_STRING( 18, URI );
    ADD_META( 19, DBUS_TYPE_INT64, i_length );

    /* Collect all ES information */
    for( int i = 0; i < p_input->i_categories ; i++)
    {
        const char *psz_cat = p_input->pp_categories[i]->psz_name;

        for( int j = 0 ; j < p_input->pp_categories[i]->i_infos ; j++ )
        {
            const char *psz_name = p_input->pp_categories[i]->pp_infos[j]->psz_name;
            const char *psz_value = p_input->pp_categories[i]->pp_infos[j]->psz_value;

            char *psz_variant = NULL;
            char *psz_entry = NULL;

            if( asprintf( &psz_entry, "%s %s", psz_cat, psz_name ) != -1 )
            {
                dbus_message_iter_open_container( &dict, DBUS_TYPE_DICT_ENTRY,
                        NULL, &dict_entry );
                dbus_message_iter_append_basic( &dict_entry, DBUS_TYPE_STRING,
                        &psz_entry );

                if( asprintf( &psz_variant, "%s", psz_value ) != -1 )
                {
                    dbus_message_iter_open_container( &dict_entry, DBUS_TYPE_VARIANT,
                            DBUS_TYPE_STRING_AS_STRING, &variant );
                    dbus_message_iter_append_basic( &variant,
                            DBUS_TYPE_STRING, &psz_variant );
                    dbus_message_iter_close_container( &dict_entry, &variant );
                    free( psz_variant );
                }
                free( psz_entry );

                dbus_message_iter_close_container( &dict, &dict_entry );
            }
        }
    }

    dbus_message_iter_close_container( args, &dict );
    return VLC_SUCCESS;
}

static int GetInputEsMeta( input_item_t * p_item, int32_t i_es,
                           DBusMessageIter *args )
{
    DBusMessageIter dict, dict_entry, variant;

    dbus_message_iter_open_container( args, DBUS_TYPE_ARRAY, "{sv}", &dict );

    if( (i_es >= 0) && (i_es < p_item->i_categories) )
    {
        for( int j = 0 ; j < p_item->pp_categories[i_es]->i_infos ; j++ )
        {
            const char *psz_name = p_item->pp_categories[i_es]->pp_infos[j]->psz_name;
            const char *psz_value = p_item->pp_categories[i_es]->pp_infos[j]->psz_value;

            char *psz_variant = NULL;

            dbus_message_iter_open_container( &dict, DBUS_TYPE_DICT_ENTRY,
                    NULL, &dict_entry );
            dbus_message_iter_append_basic( &dict_entry, DBUS_TYPE_STRING,
                    &psz_name );

            if( asprintf( &psz_variant, "%s", psz_value ) != -1 )
            {
                dbus_message_iter_open_container( &dict_entry, DBUS_TYPE_VARIANT,
                        DBUS_TYPE_STRING_AS_STRING, &variant );
                dbus_message_iter_append_basic( &variant,
                        DBUS_TYPE_STRING, &psz_variant );
                dbus_message_iter_close_container( &dict_entry, &variant );
                free( psz_variant );
            }
            dbus_message_iter_close_container( &dict, &dict_entry );

            if( !strcmp( psz_name, "Language") )
            {
                const struct iso639_lang_t *code = GetLang_iso639(psz_value);
                char *psz_code = strdup("iso639_1");

                psz_variant = NULL;

                dbus_message_iter_open_container( &dict, DBUS_TYPE_DICT_ENTRY,
                        NULL, &dict_entry );
                dbus_message_iter_append_basic( &dict_entry, DBUS_TYPE_STRING,
                        &psz_code );

                if( asprintf( &psz_variant, "%s", code->psz_iso639_1 ) != -1 )
                {
                    dbus_message_iter_open_container( &dict_entry, DBUS_TYPE_VARIANT,
                            DBUS_TYPE_STRING_AS_STRING, &variant );
                    dbus_message_iter_append_basic( &variant,
                            DBUS_TYPE_STRING, &psz_variant );
                    dbus_message_iter_close_container( &dict_entry, &variant );
                    free( psz_variant );
                }

                dbus_message_iter_close_container( &dict, &dict_entry );

                free( psz_code );

                /* */
                psz_code = strdup("iso639_2T");
                psz_variant = NULL;

                dbus_message_iter_open_container( &dict, DBUS_TYPE_DICT_ENTRY,
                        NULL, &dict_entry );
                dbus_message_iter_append_basic( &dict_entry, DBUS_TYPE_STRING,
                        &psz_code );

                if( asprintf( &psz_variant, "%s", code->psz_iso639_2T ) != -1 )
                {
                    dbus_message_iter_open_container( &dict_entry, DBUS_TYPE_VARIANT,
                            DBUS_TYPE_STRING_AS_STRING, &variant );
                    dbus_message_iter_append_basic( &variant,
                            DBUS_TYPE_STRING, &psz_variant );
                    dbus_message_iter_close_container( &dict_entry, &variant );
                    free( psz_variant );
                }

                dbus_message_iter_close_container( &dict, &dict_entry );

                free(psz_code);
            }
        }
    }

    dbus_message_iter_close_container( args, &dict );
    return VLC_SUCCESS;
}

#undef ADD_META
#undef ADD_VLC_META_STRING

/*****************************************************************************
 * MarshalStatus: Fill a DBusMessage with the current player status
 *****************************************************************************/

static int MarshalStatus( intf_thread_t* p_intf, DBusMessageIter* args,
                          bool lock )
{ /* This is NOT the right way to do that, it would be better to sore
     the status information in p_sys and update it on change, thus
     avoiding a long lock */
    VLC_UNUSED( lock );

    DBusMessageIter status;
    dbus_int32_t i_state, i_random, i_repeat, i_loop;
    vlc_value_t val;
    playlist_t* p_playlist = NULL;
    input_thread_t* p_input = NULL;

    p_playlist = pl_Hold( p_intf );

    i_state = 2;

    p_input = playlist_CurrentInput( p_playlist );
    if( p_input )
    {
        var_Get( p_input, "state", &val );
        if( val.i_int >= END_S )
            i_state = 2;
        else if( val.i_int == PAUSE_S )
            i_state = 1;
        else if( val.i_int <= PLAYING_S )
            i_state = 0;
        vlc_object_release( p_input );
    }

    i_random = var_CreateGetBool( p_playlist, "random" );

    i_repeat = var_CreateGetBool( p_playlist, "repeat" );

    i_loop = var_CreateGetBool( p_playlist, "loop" );

    pl_Release( p_intf );

    dbus_message_iter_open_container( args, DBUS_TYPE_STRUCT, NULL, &status );
    dbus_message_iter_append_basic( &status, DBUS_TYPE_INT32, &i_state );
    dbus_message_iter_append_basic( &status, DBUS_TYPE_INT32, &i_random );
    dbus_message_iter_append_basic( &status, DBUS_TYPE_INT32, &i_repeat );
    dbus_message_iter_append_basic( &status, DBUS_TYPE_INT32, &i_loop );
    dbus_message_iter_close_container( args, &status );

    return VLC_SUCCESS;
}
